unit SD;

interface

uses SysUtils, Classes, Math;

type TSampleData = class
        private
                sampleData: array of array of double;
                marginOfErrorMin : double;
                marginOfErrorMax : double;

                procedure recalculateMinMax();

        public
                rowCount: integer;
                minX, minY, maxX, maxY : double;

                constructor Create(count: integer);

                function getX(index : integer):double;
                function getY(index : integer):double;

                procedure setX(index : integer; value : double);
                procedure setY(index : integer; value : double);

                function isApproximateY(index : integer; value:double):boolean;
                function getMarginOfError():double;
                procedure setMarginOfError(value : double);

                procedure addXY(x, y: double);
                procedure remove(index:integer);

                procedure sortXY();
                procedure sortYX();

                procedure setRandomSin();
                procedure setRandom();

                procedure Clear();
                procedure Resize(count: integer);

                procedure LoadFromFile(fileName : String);
                procedure SaveToFile(fileName : String);

end;

implementation

        constructor TSampleData.Create(count: integer);
        begin
                rowCount := 0;
                resize(count);
                setMarginOfError(0.1);
        end;

        procedure TSampleData.Clear();
        var i : integer;
        begin
                minX := 0;
                minY := 0;
                maxX := 0;
                maxY := 0;

                for i := 0 to rowCount - 1 do
                begin
                        sampleData[i, 0] := 0;
                        sampleData[i, 1] := 0;
                end;
        end;

        procedure TSampleData.Resize(count: integer);
        var i : integer;
        begin
                SetLength(sampleData, count, 2);
                if rowCount < count then
                    for i := rowCount to count - 1 do
                    begin
                            sampleData[i, 0] := 0;
                            sampleData[i, 1] := 0;
                    end;
                rowCount := count;
                recalculateMinMax();
        end;

        procedure TSampleData.recalculateMinMax();
        var i : integer;
        begin
                if rowCount > 0 then
                begin
                    minX := sampleData[0, 0];
                    maxX := sampleData[0, 0];
                    minY := sampleData[0, 1];
                    maxY := sampleData[0, 1];

                    for i := 1 to rowCount - 1 do
                    begin
                        if sampleData[i, 0] > maxX then
                           maxX := sampleData[i, 0];
                        if sampleData[i, 0] < minX then
                           minX := sampleData[i, 0];
                        if sampleData[i, 1] > maxY then
                           maxY := sampleData[i, 1];
                        if sampleData[i, 1] < minY then
                           minY := sampleData[i, 1];
                    end;
                end
                else
                begin
                    minX := 0;
                    minY := 0;
                    maxX := 0;
                    maxY := 0;
                end;
        end;

        function TSampleData.getX(index : integer):double;
        begin
                result := sampleData[index, 0];
        end;

        function TSampleData.getY(index : integer):double;
        begin
                result := sampleData[index, 1];
        end;

        procedure TSampleData.setX(index : integer; value : double);
        begin
                if value > maxX then
                   maxX := value;
                if value < minX then
                   minX := value;
                sampleData[index, 0] := value;
        end;

        procedure TSampleData.setY(index : integer; value : double);
        begin
                if value > maxY then
                   maxY := value;
                if value < minY then
                   minY := value;
                sampleData[index, 1] := value;
        end;

        function TSampleData.isApproximateY(index : integer; value:double):boolean;
        var ok : boolean;
        begin
                ok := false;

                if (value >= minY) and (value <= maxY) then
                begin
                        if value = sampleData[index, 1] then
                                ok := true
                        else
                        begin
                                if value < sampleData[index, 1] then
                                begin
                                        if ((value - minY) / (sampleData[index, 1] - minY)) >= marginOfErrorMax then
                                                ok := true;
                                end
                                else
                                begin
                                        if ((value - sampleData[index, 1]) / (maxY - sampleData[index, 1])) <= marginOfErrorMin then
                                                ok := true;
                                end;
                        end;
                end;

                result := ok;
        end;

        function TSampleData.getMarginOfError():double;
        begin
                result := marginOfErrorMin;
        end;

        procedure TSampleData.setMarginOfError(value : double);
        begin
                marginOfErrorMin := value;
                marginOfErrorMax := 1 - marginOfErrorMin;
        end;

        procedure TSampleData.addXY(x, y: double);
        begin
                resize(rowCount+1);
                sampleData[rowCount-1, 0] := x;
                sampleData[rowCount-1, 1] := y;
                if x > maxX then
                   maxX := x;
                if x < minX then
                   minX := x;
                if y > maxY then
                   maxY := y;
                if y < minY then
                   minY := y;
        end;

        procedure TSampleData.remove(index:integer);
        var i : integer;
        begin
                for i := index to rowCount - 2 do
                begin
                        sampleData[i, 0] := sampleData[i+1, 0];
                        sampleData[i, 1] := sampleData[i+1, 1];
                end;
                Resize(rowCount - 1);
        end;

        procedure TSampleData.sortXY();
        var i, k : integer;
            t : double;
        begin
                for i := 0 to rowCount - 2 do
                    for k := i+1 to rowCount - 1 do
                    begin
                        if sampleData[i, 0] > sampleData[k, 0] then
                        begin
                            t := sampleData[i, 0];
                            sampleData[i, 0] := sampleData[k, 0];
                            sampleData[k, 0] := t;
                            t := sampleData[i, 1];
                            sampleData[i, 1] := sampleData[k, 1];
                            sampleData[k, 1] := t;
                        end
                        else
                        if sampleData[i, 0] = sampleData[k, 0] then
                        begin
                                if sampleData[i, 1] > sampleData[k, 1] then
                                begin
                                    t := sampleData[i, 1];
                                    sampleData[i, 1] := sampleData[k, 1];
                                    sampleData[k, 1] := t;
                                end;
                        end;
                    end;

        end;

        procedure TSampleData.sortYX();
        var i, k : integer;
            t : double;
        begin
                for i := 0 to rowCount - 2 do
                    for k := i+1 to rowCount - 1 do
                    begin
                        if sampleData[i, 1] > sampleData[k, 1] then
                        begin
                            t := sampleData[i, 0];
                            sampleData[i, 0] := sampleData[k, 0];
                            sampleData[k, 0] := t;
                            t := sampleData[i, 1];
                            sampleData[i, 1] := sampleData[k, 1];
                            sampleData[k, 1] := t;
                        end
                        else
                        if sampleData[i, 1] = sampleData[k, 1] then
                        begin
                                if sampleData[i, 0] > sampleData[k, 0] then
                                begin
                                    t := sampleData[i, 0];
                                    sampleData[i, 0] := sampleData[k, 0];
                                    sampleData[k, 0] := t;
                                end;
                        end;
                    end;
        end;

        procedure TSampleData.setRandomSin();
        var i : integer;
            c : double;
        begin
              c := random(50)+1;

              for i := 0 to rowCount - 1 do
              begin
                  sampleData[i, 0] := i;
                  sampleData[i, 1] := sin(c*i);
              end;

              recalculateMinMax();
        end;

        procedure TSampleData.setRandom();
        var i : integer;
        begin
              for i := 0 to rowCount - 1 do
              begin
                  sampleData[i, 0] := random;
                  sampleData[i, 1] := random;
              end;
              recalculateMinMax();
              sortXY();
        end;

        procedure TSampleData.LoadFromFile(fileName : String);
        var s1, s2:TStringList;
            i : integer;
            x, y : double;
        begin
              s1 := TStringList.Create;
              s2 := TStringList.Create;
              s1.Delimiter := chr(13);
              s2.Delimiter := ',';
              s1.LoadFromFile(fileName);

              resize(s1.Count);

              for i := 0 to rowCount - 1 do
              begin
                  s2.DelimitedText := s1.Strings[i];
                  try
                        x := StrToFloat(s2.Strings[0]);
                        y := StrToFloat(s2.Strings[1]);
                  except
                        x := 0;
                        y := 0;
                  end;
                  setX(i, x);
                  setY(i, y);
              end;

              s2.free;
              s1.Free;
        end;

        procedure TSampleData.SaveToFile(fileName : String);
        var s1, s2:TStringList;
            i : integer;
            x, y : double;
        begin
              s1 := TStringList.Create;
              s2 := TStringList.Create;
              s1.Delimiter := chr(13);
              s2.Delimiter := ',';
              s2.Add('1');
              s2.Add('2');

              for i := 0 to rowCount - 1 do
              begin
                  s2.Strings[0] := FloatToStrF(getX(i), ffFixed, 16, 8);
                  s2.Strings[1] := FloatToStrF(getY(i), ffFixed, 16, 8);
                  s1.Add(s2.DelimitedText);
              end;

              s1.SaveToFile(fileName);

              s2.free;
              s1.Free;
        end;

end.
