unit TE;

interface

uses SysUtils, Classes, Math;

const maxOperatorCode = 13;
        constantParm = 255;

type TTuringCode = record
        parm1 : byte;
        operator : byte;
        parm2 : byte;
        constant : double;
end;

type TTuringEvaluator = class
        private
                instructionCount, varsCount: integer;
                instructionData: array of TTuringCode;
                validCount : integer;
                executeNext : boolean;
                v : array [0..19] of double;

                procedure calcValidCount();

        public

                constructor Create(count, vars:integer);
                procedure randomPopulate();
                function Evaluate(value : double):double;

                procedure setCrossover(from : TTuringEvaluator);

                procedure setData(index: integer; data : TTuringCode);
                procedure setDataRandom(index: integer);
                function getData(index: integer):TTuringCode;
                function getCodeText():String;
end;

implementation

        constructor TTuringEvaluator.Create(count, vars:integer);
        begin
                if vars < 1 then
                        vars := 1
                else if vars > 20 then
                        vars := 20;

                instructionCount := count;
                varsCount := vars;
                validCount := 0;

                setLength(instructionData, count);
        end;

        procedure TTuringEvaluator.randomPopulate();
        var i : integer;
        begin
                for i := 0 to instructionCount - 1 do
                begin
                        setDataRandom(i);
                end;
                calcValidCount;
        end;

        procedure TTuringEvaluator.setCrossover(from : TTuringEvaluator);
        var i : integer;
        begin
                for i := 0 to instructionCount - 1 do
                begin
                    if random(100) < 50 then  // metade do pai, metade da me
                       instructionData[i] := from.getData(i);
                end;
                calcValidCount;
        end;

        procedure TTuringEvaluator.setData(index: integer; data : TTuringCode);
        begin
                instructionData[index] := data;
        end;

        function TTuringEvaluator.getData(index: integer):TTuringCode;
        begin
                result := instructionData[index];
        end;

        procedure TTuringEvaluator.setDataRandom(index: integer);
        begin
                instructionData[index].parm1 := random(varsCount);
                instructionData[index].operator := random(maxOperatorCode);
                if (random(100) > 50) or ((instructionData[index].operator >= 5) and (instructionData[index].operator<=10)) then
                        instructionData[index].parm2 := random(varsCount)
                else
                begin
                        instructionData[index].parm2 := constantParm;
                        instructionData[index].constant := random(99999999)/10000.0;
                end;
        end;

        procedure TTuringEvaluator.calcValidCount();
        var i : integer;
        begin
                validCount := instructionCount - 1;
                for i := instructionCount - 1 downto 0 do
                begin
                        if (instructionData[i].parm1 = 0) then
                        begin
                                validCount := i;
                                break;
                        end;
                end;
        end;

        function TTuringEvaluator.Evaluate(value : double):double;
        var i : integer;
        begin
                executeNext := true;
                
                for i := 1 to varsCount do
                        v[i] := 0;

                v[0] := value;

                for i := 0 to validCount do
                begin
                        if(executeNext) then
                        begin
                            try
                                case instructionData[i].operator of
                                0 : // +=
                                        begin
                                                if instructionData[i].parm2 = constantParm then
                                                        v[instructionData[i].parm1] := v[instructionData[i].parm1] + instructionData[i].constant
                                                else
                                                        v[instructionData[i].parm1] := v[instructionData[i].parm1] + v[instructionData[i].parm2];
                                        end;
                                1 : // -=
                                        begin
                                                if instructionData[i].parm2 = constantParm then
                                                        v[instructionData[i].parm1] := v[instructionData[i].parm1] - instructionData[i].constant
                                                else
                                                        v[instructionData[i].parm1] := v[instructionData[i].parm1] - v[instructionData[i].parm2];
                                        end;
                                2 : // *=
                                        begin
                                                if instructionData[i].parm2 = constantParm then
                                                        v[instructionData[i].parm1] := v[instructionData[i].parm1] * instructionData[i].constant
                                                else
                                                        v[instructionData[i].parm1] := v[instructionData[i].parm1] * v[instructionData[i].parm2];
                                        end;
                                3 : // /=
                                        begin
                                                if instructionData[i].parm2 = constantParm then
                                                begin
                                                        if instructionData[i].constant <> 0 then
                                                                v[instructionData[i].parm1] := v[instructionData[i].parm1] / instructionData[i].constant;
                                                end
                                                else
                                                begin
                                                        if v[instructionData[i].parm2] <> 0 then
                                                                v[instructionData[i].parm1] := v[instructionData[i].parm1] / v[instructionData[i].parm2];
                                                end;
                                        end;
                                4 : // ^=
                                        begin
                                                if instructionData[i].parm2 = constantParm then
                                                begin
                                                        if(instructionData[i].constant <= 10) then
                                                              v[instructionData[i].parm1] := power(v[instructionData[i].parm1], instructionData[i].constant)
                                                end
                                                else
                                                begin
                                                        if(v[instructionData[i].parm2] <= 10) then
                                                              v[instructionData[i].parm1] := power(v[instructionData[i].parm1], v[instructionData[i].parm2]);
                                                end
                                        end;
                                5 : // sin
                                        begin
                                                if instructionData[i].parm2 = constantParm then
                                                        v[instructionData[i].parm1] := sin(instructionData[i].constant)
                                                else
                                                        v[instructionData[i].parm1] := sin(v[instructionData[i].parm2]);
                                        end;
                                6 : // cos
                                        begin
                                                if instructionData[i].parm2 = constantParm then
                                                        v[instructionData[i].parm1] := cos(instructionData[i].constant)
                                                else
                                                        v[instructionData[i].parm1] := cos(v[instructionData[i].parm2]);
                                        end;
                                7 : // tan
                                        begin
                                                if instructionData[i].parm2 = constantParm then
                                                        v[instructionData[i].parm1] := tan(instructionData[i].constant)
                                                else
                                                        v[instructionData[i].parm1] := tan(v[instructionData[i].parm2]);
                                        end;
                                8 : // exp
                                        begin
                                                if instructionData[i].parm2 = constantParm then
                                                begin
                                                      if (instructionData[i].constant >= 0) and (instructionData[i].constant < 1) then
                                                              v[instructionData[i].parm1] := exp(instructionData[i].constant);
                                                end
                                                else
                                                begin
                                                      if (v[instructionData[i].parm2] >= 0) and (v[instructionData[i].parm2] < 1) then
                                                              v[instructionData[i].parm1] := exp(v[instructionData[i].parm2]);
                                                end;
                                        end;
                                9 : // sqrt
                                        begin
                                                if instructionData[i].parm2 = constantParm then
                                                begin
                                                        if instructionData[i].constant >= 0 then
                                                                v[instructionData[i].parm1] := sqrt(instructionData[i].constant);
                                                end
                                                else
                                                begin
                                                        if v[instructionData[i].parm2] >= 0 then
                                                                v[instructionData[i].parm1] := sqrt(v[instructionData[i].parm2]);
                                                end;
                                        end;
                                10 : // ln
                                        begin
                                                if instructionData[i].parm2 = constantParm then
                                                begin
                                                        if instructionData[i].constant <> 0 then
                                                                v[instructionData[i].parm1] := ln(instructionData[i].constant);
                                                end
                                                else
                                                begin
                                                        if v[instructionData[i].parm2] <> 0 then
                                                                v[instructionData[i].parm1] := ln(v[instructionData[i].parm2]);
                                                end;
                                        end;
                                11 : // if(parm1 > parm2)
                                        begin
                                                if instructionData[i].parm2 = constantParm then
                                                begin
                                                        executeNext := (v[instructionData[i].parm1] > instructionData[i].constant);
                                                end
                                                else
                                                begin
                                                        executeNext := (v[instructionData[i].parm1] > v[instructionData[i].parm2]);
                                                end;
                                        end;
                                12 : // if(parm1 <= parm2)
                                        begin
                                                if instructionData[i].parm2 = constantParm then
                                                begin
                                                        executeNext := (v[instructionData[i].parm1] <= instructionData[i].constant);
                                                end
                                                else
                                                begin
                                                        executeNext := (v[instructionData[i].parm1] <= v[instructionData[i].parm2]);
                                                end;
                                        end;
                                else // variables
                                        begin
                                                if instructionData[i].parm2 = constantParm then
                                                        v[instructionData[i].parm1] := instructionData[i].constant
                                                else
                                                        v[instructionData[i].parm1] := v[instructionData[i].parm2];
                                        end;
                                end;
                            except
                            end;
                        end
                        else
                        begin
                                // avoid nested ifs...
                                if(instructionData[i].operator <> 11) and (instructionData[i].operator <> 12) then
                                        executeNext := true;
                        end;

                end;

                result := v[0];
        end;

        function TTuringEvaluator.getCodeText():String;
        var i, maxv : integer;
            s, crlf : string;
            insertCRLF:boolean;
            fs : TFormatSettings;
        begin
                crlf := chr(13) + chr(10);
                insertCRLF := true;

                GetLocaleFormatSettings($0409, fs);

                s := 'double model(double valueX) {' + crlf;

                maxv := 0;
                for i := 0 to validCount do
                begin
                        if instructionData[i].parm1 > maxv then
                                maxv := instructionData[i].parm1;
                        if instructionData[i].parm2 <> constantParm then
                                if instructionData[i].parm2 > maxv then
                                        maxv := instructionData[i].parm2;
                end;

                s := s + 'double var[' + IntToStr(maxv+1) + '];' + crlf;
                s := s + 'var[0] = valueX;' + crlf;

                for i := 1 to maxv do
                        s := s + 'var['+IntToStr(i)+'] = 0;' + chr(13) + chr(10);

                for i := 0 to validCount do
                begin

                        insertCRLF := true;

                        case instructionData[i].operator of
                        0 : // +=
                                begin
                                        if instructionData[i].parm2 = constantParm then
                                                s := s + 'var[' + IntToStr(instructionData[i].parm1) + '] += ' + FloatToStrF(instructionData[i].constant, ffFixed, 15, 4, fs)
                                        else
                                                s := s + 'var[' + IntToStr(instructionData[i].parm1) + '] += var[' + IntToStr(instructionData[i].parm2) + ']';
                                end;
                        1 : // -=
                                begin
                                        if instructionData[i].parm2 = constantParm then
                                                s := s + 'var[' + IntToStr(instructionData[i].parm1) + '] -= ' + FloatToStrF(instructionData[i].constant, ffFixed, 15, 4, fs)
                                        else
                                                s := s + 'var[' + IntToStr(instructionData[i].parm1) + '] -= var[' + IntToStr(instructionData[i].parm2) + ']';
                                end;
                        2 : // *=
                                begin
                                        if instructionData[i].parm2 = constantParm then
                                                s := s + 'var[' + IntToStr(instructionData[i].parm1) + '] *= ' + FloatToStrF(instructionData[i].constant, ffFixed, 15, 4, fs)
                                        else
                                                s := s + 'var[' + IntToStr(instructionData[i].parm1) + '] *= var[' + IntToStr(instructionData[i].parm2) + ']';
                                end;
                        3 : // /=
                                begin
                                        if instructionData[i].parm2 = constantParm then
                                                s := s + 'var[' + IntToStr(instructionData[i].parm1) + '] /= ' + FloatToStrF(instructionData[i].constant, ffFixed, 15, 4, fs)
                                        else
                                                s := s + 'var[' + IntToStr(instructionData[i].parm1) + '] /= var[' + IntToStr(instructionData[i].parm2) + ']';
                                end;
                        4 : // ^=
                                begin
                                        if instructionData[i].parm2 = constantParm then
                                                s := s + 'var[' + IntToStr(instructionData[i].parm1) + '] ^= ' + FloatToStrF(instructionData[i].constant, ffFixed, 15, 4, fs)
                                        else
                                                s := s + 'var[' + IntToStr(instructionData[i].parm1) + '] ^= var[' + IntToStr(instructionData[i].parm2) + ']';
                                end;
                        5 : // sin
                                begin
                                        if instructionData[i].parm2 = constantParm then
                                                s := s + 'var[' + IntToStr(instructionData[i].parm1) + '] = SIN(' + FloatToStrF(instructionData[i].constant, ffFixed, 15, 4, fs) + ')'
                                        else
                                                s := s + 'var[' + IntToStr(instructionData[i].parm1) + '] = SIN(var[' + IntToStr(instructionData[i].parm2) + '])';
                                end;
                        6 : // cos
                                begin
                                        if instructionData[i].parm2 = constantParm then
                                                s := s + 'var[' + IntToStr(instructionData[i].parm1) + '] = COS(' + FloatToStrF(instructionData[i].constant, ffFixed, 15, 4, fs) + ')'
                                        else
                                                s := s + 'var[' + IntToStr(instructionData[i].parm1) + '] = COS(var[' + IntToStr(instructionData[i].parm2) + '])';
                                end;
                        7 : // tan
                                begin
                                        if instructionData[i].parm2 = constantParm then
                                                s := s + 'var[' + IntToStr(instructionData[i].parm1) + '] = TAN(' + FloatToStrF(instructionData[i].constant, ffFixed, 15, 4, fs) + ')'
                                        else
                                                s := s + 'var[' + IntToStr(instructionData[i].parm1) + '] = TAN(var[' + IntToStr(instructionData[i].parm2) + '])';
                                end;
                        8 : // exp
                                begin
                                        if instructionData[i].parm2 = constantParm then
                                                s := s + 'var[' + IntToStr(instructionData[i].parm1) + '] = EXP(' + FloatToStrF(instructionData[i].constant, ffFixed, 15, 4, fs) + ')'
                                        else
                                                s := s + 'var[' + IntToStr(instructionData[i].parm1) + '] = EXP(var[' + IntToStr(instructionData[i].parm2) + '])';
                                end;
                        9 : // sqrt
                                begin
                                        if instructionData[i].parm2 = constantParm then
                                                s := s + 'var[' + IntToStr(instructionData[i].parm1) + '] = SQRT(' + FloatToStrF(instructionData[i].constant, ffFixed, 15, 4, fs) + ')'
                                        else
                                                s := s + 'var[' + IntToStr(instructionData[i].parm1) + '] = SQRT(var[' + IntToStr(instructionData[i].parm2) + '])';
                                end;
                        10 : // ln
                                begin
                                        if instructionData[i].parm2 = constantParm then
                                                s := s + 'var[' + IntToStr(instructionData[i].parm1) + '] = LN(' + FloatToStrF(instructionData[i].constant, ffFixed, 15, 4, fs) + ')'
                                        else
                                                s := s + 'var[' + IntToStr(instructionData[i].parm1) + '] = LN(var[' + IntToStr(instructionData[i].parm2) + '])';
                                end;
                        11 : // if(parm1 > parm2)
                                begin
                                        insertCRLF := false;
                                        if instructionData[i].parm2 = constantParm then
                                                s := s + 'if(var[' + IntToStr(instructionData[i].parm1) + '] > ' + FloatToStrF(instructionData[i].constant, ffFixed, 15, 4, fs) + ') '
                                        else
                                                s := s + 'if(var[' + IntToStr(instructionData[i].parm1) + '] > var[' + IntToStr(instructionData[i].parm2) + ']) ';
                                end;
                        12 : // if(parm1 <= parm2)
                                begin
                                        insertCRLF := false;
                                        if instructionData[i].parm2 = constantParm then
                                                s := s + 'if(var[' + IntToStr(instructionData[i].parm1) + '] <= ' + FloatToStrF(instructionData[i].constant, ffFixed, 15, 4, fs) + ') '
                                        else
                                                s := s + 'if(var[' + IntToStr(instructionData[i].parm1) + '] <= var[' + IntToStr(instructionData[i].parm2) + ']) ';
                                end;
                        else // variables
                                begin
                                        if instructionData[i].parm2 = constantParm then
                                                s := s + 'var[' + IntToStr(instructionData[i].parm1) + '] = ' + FloatToStrF(instructionData[i].constant, ffFixed, 15, 4, fs)
                                        else
                                                s := s + 'var[' + IntToStr(instructionData[i].parm1) + '] = var[' + IntToStr(instructionData[i].parm2) + ']';
                                end;
                        end;

                        if(insertCRLF) then
                                s := s + ';' + crlf;
                end;

                if(not insertCRLF) then
                        s := s + 'var[0] = var[0];' + crlf;

                s := s + 'return var[0];' + crlf + '}';

                result := s;
        end;

end.
